/*
 *
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.nextian.icmpping.helpers;

import com.nextian.icmpping.helpers.exceptions.PingErrorException;
import com.nextian.icmpping.helpers.handlers.BlockingPing;
import com.nextian.icmpping.helpers.handlers.PingRequest;
import com.nextian.icmpping.helpers.handlers.Receiver;
import com.nextian.icmpping.helpers.handlers.Sender;
import com.nextian.icmpping.jni.IcmpSocket;
import com.nextian.icmpping.jni.exception.IcmpSystemException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.InetAddress;

/**
 * A wrapper class around static methods providing ICMP echo (ping) functionality.
 * Once {@link #init()} as been called, two threads (sender and receiver) are created for sending queued requests
 * and processing responses. Use {@link #release} to stop the threads and release all resources.
 */
public class Ping {
    private static final Logger LOGGER = LoggerFactory.getLogger(Ping.class);

    private static IcmpSocket icmpSocket;
    private static Sender sender;
    private static Receiver receiver;
    private static Thread senderThread;
    private static Thread receiverThread;

    /**
     * Private constructor, preventing class from being instantiated (it has only static method and should
     * be treated as a namespace.
     */
    private Ping() {
    }

    /**
     * Initialize members: creates raw ICMP socket and creates sender and receiver threads.
     * Call {@link #release} to release allocated resources.
     *
     * @throws PingErrorException when resources cannot be allocated or init() has already been called
     */
    static public synchronized void init() throws PingErrorException {
        if (icmpSocket == null) {
            try {
                icmpSocket = new IcmpSocket();
                icmpSocket.open();
                initThreads();
            } catch (IcmpSystemException e) {
                icmpSocket = null;
                throw new PingErrorException("Cannot open ICMP socket", e);
            }
        } else {
            throw new PingErrorException("Already initialized", null);
        }
    }

    /**
     * Initialize sender and receiver threads.
     *
     * @throws PingErrorException when threads cannot be started
     */
    static private void initThreads() throws PingErrorException {
        try {
            receiver = new Receiver(icmpSocket);
            sender = new Sender(receiver, icmpSocket);
            receiverThread = new Thread(receiver);
            senderThread = new Thread(sender);
            receiverThread.start();
            senderThread.start();
        } catch (Exception e) {
            release();
            throw new PingErrorException("Cannot initialize PingManager", e);
        }
    }

    /**
     * Release resources allocated in {@link #init} method.
     * Internal exceptions are either properly handled or logged using {@link Logger}.
     */
    public static synchronized void release() {
        try {
            if (icmpSocket != null) {
                icmpSocket.close();
                icmpSocket = null; // null assigned to accommodate multiple init()/release() sequences
            }
        } catch (IcmpSystemException e) {
            LOGGER.error("Socket close failed", e);
        }

        try {
            if (senderThread != null) {
                senderThread.interrupt();
                senderThread.join();
            }
            if (receiverThread != null) {
                receiverThread.interrupt();
                receiverThread.join();
            }
        } catch (InterruptedException e) {
            LOGGER.error("Could not stop threads", e);
        }
    }

    /**
     * Asynchronous ICMP echo request with user provided echo message data.
     *
     * @param destination pinged host address
     * @param data        data to be send as part of ICMP message
     * @param timeout     millisecond timeout to wait for a response
     * @param handler     response handler
     * @return unique ICMP message id identifying the request
     * @throws PingErrorException issued if Ping class is not initialized
     */
    public static synchronized int echo(InetAddress destination, byte[] data, int timeout, PingResponseHandler handler) throws PingErrorException {
        int messageId;

        if (icmpSocket != null) {
            messageId = icmpSocket.getNextMessageId();
            sender.registerRequest(new PingRequest(destination, data, data.length, timeout, messageId, handler));
        } else {
            throw new PingErrorException("Not initialized", null);
        }
        return messageId;
    }

    /**
     * Asynchronous ICMP echo request with user provided echo message data.
     *
     * @param destination pinged host address
     * @param dataLength  number of bytes to send in the echo message
     * @param timeout     millisecond timeout to wait for a response
     * @param handler     response handler
     * @return unique ICMP message id identifying the request
     * @throws PingErrorException issued if Ping class is not initialized
     */
    public static synchronized int echo(InetAddress destination, int dataLength, int timeout, PingResponseHandler handler) throws PingErrorException {
        int messageId;

        if (sender != null) {
            messageId = icmpSocket.getNextMessageId();
            sender.registerRequest(new PingRequest(destination, null, dataLength, timeout, messageId, handler));
        } else {
            throw new PingErrorException("Ping not initialized", null);
        }
        return messageId;
    }

    /**
     * Send a set of echo requests with specified message data. Method is executed synchronously and waits until all echo requests
     * are completed.
     *
     * @param addresses  array addresses to ping
     * @param dataToSend data to be send as part of ICMP message
     * @param timeout    milliseconds time to wait for echo responses
     * @return array of ping results
     * @throws PingErrorException   Ping class is not initialized
     * @throws InterruptedException execution interrupted
     */
    public static PingResult[] echo(InetAddress[] addresses, byte[] dataToSend, int timeout) throws PingErrorException, InterruptedException {
        BlockingPing ping = new BlockingPing();
        return ping.echo(addresses, dataToSend, dataToSend.length, timeout);
    }

    /**
     * Send a set of echo requests with default message data. Method is executed synchronously and waits until all echo requests
     * are completed.
     *
     * @param addresses  array addresses to ping
     * @param dataLength number of bytes to send in the echo message
     * @param timeout    milliseconds time to wait for echo responses
     * @return array of ping results
     * @throws PingErrorException   Ping class is not initialized
     * @throws InterruptedException execution interrupted
     */
    public static PingResult[] echo(InetAddress[] addresses, int dataLength, int timeout) throws PingErrorException, InterruptedException {
        BlockingPing ping = new BlockingPing();
        return ping.echo(addresses, null, dataLength, timeout);
    }
}
