/*
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 *
 */
package com.nextian.icmpping.helpers;

import com.nextian.icmpping.IcmpStatus;

import java.net.InetAddress;

/**
 * The class is a definition of results returned by synchronous version of {@link Ping#echo} method.
 */
public class PingResult {

    private final InetAddress address;
    private final IcmpStatus status;
    private final ResultType result;
    private final long sentTime;
    private final long responseTime;

    /**
     * Class constructor.
     *
     * @param address  the address of ICMP EchoRequest destination host
     * @param status   status of ICMP response
     * @param result   type of result
     * @param sentTime time when request echo was sent //TimeUtil.currentTimeMillis();
     */
    public PingResult(InetAddress address, IcmpStatus status, ResultType result, long sentTime, long responseTime) {
        this.address = address;
        this.status = status;
        this.result = result;
        this.sentTime = sentTime;
        this.responseTime = responseTime;
    }

    /**
     * The {@link PingResult} is response to ICMP EchoRequest.
     * This method get address of host the ICMP EchoRequest was sent to.
     *
     * @return address of host request was sent to
     */
    public InetAddress getAddress() {
        return address;
    }

    /**
     * Get time of request processing. It is difference between response receive time and
     * request send time.
     *
     * @return echo processing time
     */
    public long getResponseTime() {
        return responseTime - sentTime;
    }

    /**
     * Get type of result content.
     *
     * @return result type
     */
    public ResultType getResult() {
        return result;
    }

    /**
     * Get ICMP status. The received response may be either successful EchoResponse or other
     * kind of message. The status describes its type.
     *
     * @return status of ICMP response
     */
    public IcmpStatus getStatus() {
        return status;
    }

    /**
     * Enum defines type of result
     */
    public enum ResultType {
        /**
         * Result contains ICMP response
         */
        RESPONSE(0),
        /**
         * TIMEOUT occurred when receiving ICMP request response
         */
        TIMEOUT(1),
        /**
         * ERROR in receiving data
         */
        ERROR(2);

        private final int value;

        ResultType(int code) {
            value = code;
        }

        /**
         * Get integer code of enum
         *
         * @return integer code of enum
         */
        public int getValue() {
            return value;
        }
    }
}
