/*
 *
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.nextian.icmpping.helpers.examples;

import com.nextian.icmpping.helpers.Ping;
import com.nextian.icmpping.helpers.PingResponseHandler;
import com.nextian.icmpping.helpers.exceptions.PingErrorException;
import com.nextian.icmpping.helpers.handlers.PingRequest;
import com.nextian.icmpping.jni.IcmpEchoResponse;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.concurrent.CountDownLatch;

/**
 * Asynchronous command line helpers.
 */
public class PingExampleAsync {
    /**
     * Convert list of DNS host names into IP addresses. Invalid (non-resolvable) addresses are skipped.
     *
     * @param args DNS host names
     * @return resolved IP host addresses
     */
    private static InetAddress[] getAddresses(String[] args) {
        InetAddress address;
        InetAddress[] addresses;
        ArrayList<InetAddress> list = new ArrayList<InetAddress>();

        for (int i = 1; i < args.length; i++) {
            try {
                address = Inet4Address.getByName(args[i]);
                if (address != null) {
                    list.add(address);
                } else {
                    System.err.println("Unknown host: " + args[i]);
                }
            } catch (UnknownHostException ex) {
                System.err.println("Unknown host: " + args[i]);
            }
        }
        addresses = list.toArray(new InetAddress[list.size()]);
        return addresses;
    }

    /**
     * Ping all hosts asynchronously and wait for responses.
     *
     * @param addresses list of hosts addresses to helpers
     * @param countDown counter to decrease on helpers response
     * @param timeout   helpers timeout (milliseconds)
     */
    private static void parallelPing(InetAddress[] addresses, CountDownLatch countDown, int timeout) {
        // Initialize counting handler for helpers responses
        Handler handler = new Handler(countDown);

        try {
            // Send all requests
            for (InetAddress address : addresses) {
                Ping.echo(address, "PING DATA".getBytes(), timeout, handler);
            }
        } catch (PingErrorException ex) {
            System.err.println("Fatal error - terminating " + ex.toString());
        }
        try {
            countDown.await();
        } catch (InterruptedException ex) {
            System.err.println("Program terminated, exiting");
        }
    }

    /**
     * Execute parallel (made asynchronously) helpers request
     *
     * @param args list of parameters containing timeout and list of host names
     */
    public static void main(String[] args) {
        if (args.length >= 2) {
            int timeout = Integer.parseInt(args[0]);
            InetAddress[] addresses = getAddresses(args);
            CountDownLatch countDownLatch = new CountDownLatch(addresses.length);

            try {
                Ping.init();
                parallelPing(addresses, countDownLatch, timeout);
                countDownLatch.await();
            } catch (PingErrorException ex) {
                System.err.println("Could not initialize helpers: " + ex.toString());
            } catch (InterruptedException e) {
                // Nothing to do - program is exiting
            } finally {
                Ping.release();
            }
        } else {
            System.out.println("Usage:");
            System.out.println("     Ping <timeout> <address list>");
            System.out.println("");
        }
    }

    /**
     * A handler class for dispatching responses {@link com.nextian.icmpping.helpers.PingResponseHandler}.
     */
    private static class Handler implements PingResponseHandler {
        /**
         * A latch used to wait for all responses to come (or time out). Every handler callback decreases
         * count on the latch.
         */
        private final CountDownLatch countDownLatch;

        /**
         * Constructor, initialize {@link #countDownLatch} to number of helpers requests sent.
         *
         * @param pingCount number of helpers requests
         */
        Handler(CountDownLatch pingCount) {
            countDownLatch = pingCount;
        }

        public synchronized void handleResponse(PingRequest request, IcmpEchoResponse response) {
            System.out.println(
                    new StringBuilder()
                            .append(request.getAddress().getHostAddress())
                            .append(" from ")
                            .append(response.getSourceAddress().getHostAddress())
                            .append(" response: ")
                            .append(response.getStatus())
                            .append(" time:")
                            .append(response.getResponseTime() - request.getSendTime())
                            .append(" ms")
                            .toString());
            countDownLatch.countDown();
        }

        public synchronized void handleTimeout(PingRequest request) {
            System.out.println(request.getAddress().getHostAddress() + " timeout");
            countDownLatch.countDown();
        }

        public synchronized void handleException(PingRequest request, Exception exception) {
            exception.printStackTrace();
            System.out.println(request.getAddress().getHostAddress() + "helpers error " + exception.toString());
            countDownLatch.countDown();
        }
    }
}
