/*
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 *
 */
package com.nextian.icmpping.helpers.handlers;

import com.nextian.icmpping.IcmpStatus;
import com.nextian.icmpping.helpers.Ping;
import com.nextian.icmpping.helpers.PingResponseHandler;
import com.nextian.icmpping.helpers.PingResult;
import com.nextian.icmpping.helpers.exceptions.PingErrorException;
import com.nextian.icmpping.jni.IcmpEchoResponse;
import com.nextian.icmpping.util.TimeUtil;

import java.net.InetAddress;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.CountDownLatch;

/**
 * Implementation of {@link PingResponseHandler} for synchronous processing of ICMP echo requests
 * using asynchronous {@link Ping#echo} method.
 */
public class BlockingPing implements PingResponseHandler {
    private PingResult[] result;
    private Map<Integer, Integer> requestsMap;
    private CountDownLatch countDown;

    /*
     * Assign result to corresponding array entry. The array index is retrieved by <code>messageId</code>
     * from index map created when requests were sent.
     * The CountDownLatch counter is decreased so on last result the main thread (blocked
     * echo method) may continue.
     */
    private void setResult(int messageId, PingResult pingResult) {
        synchronized (requestsMap) {
            Integer resultIndex = requestsMap.get(messageId);
            // In general it is impossible to have "null" here
            if (resultIndex != null) {
                result[resultIndex] = pingResult;
            }
            countDown.countDown();
        }
    }

    public void handleResponse(PingRequest request, IcmpEchoResponse response) {
        setResult(request.getMessageId(), new PingResult(request.getAddress(), response.getStatus(), PingResult.ResultType.RESPONSE, request.getSendTime(), response.getResponseTime()));
    }

    public void handleTimeout(PingRequest request) {
        setResult(request.getMessageId(), new PingResult(request.getAddress(), IcmpStatus.UnexpectedResponse, PingResult.ResultType.TIMEOUT, request.getSendTime(), TimeUtil.currentTimeMillis()));
    }

    public void handleException(PingRequest request, Exception exception) {
        setResult(request.getMessageId(), new PingResult(request.getAddress(), IcmpStatus.UnexpectedResponse, PingResult.ResultType.ERROR, request.getSendTime(), TimeUtil.currentTimeMillis()));
    }

    /**
     * Send the set of ICMP echo requests. Method is executed synchronously and waits until all echo requests
     * are finished. Internally, echo requests are send asynchronously - it means all requests are send immediately
     * without waiting for responses (those, are handled in separate thread). Next, method waits until either all echo
     * responses are received or timeouts occurs.
     * <p>
     * Each echo result is placed in array of results corresponding to <code>addresses</code> parameter (it means for each input host
     * the result is placed in corresponding element of result array).
     * <p>
     * Each request timeout is calculated separately - it means that its calculation starts in the moment
     * when request is sent (not when <code>echo</code> method is called). Such behavior leads to
     * situation that <code>echo</code> execution lasts slightly longer than timeout specified.
     * <p>
     * Method reports exceptions when either {@link Ping} is not initialized or current thread is interrupted.
     *
     * @param addresses  array of host addresses send ICMP echo requests
     * @param dataToSend data to be send as part of ICMP message (when null the default data is used)
     * @param dataLength length of data to be sent
     * @param timeout    expressed in milliseconds time to wait for echo response
     * @return the array of results
     * @throws PingErrorException issued if Ping class is not initialized or thread is interrupted
     */
    public PingResult[] echo(InetAddress[] addresses, byte[] dataToSend, int dataLength, int timeout) throws PingErrorException, InterruptedException {
        if (addresses != null) {
            try {
                // Initialize waiting counter so program can wait until all requests are handled
                countDown = new CountDownLatch(addresses.length);
                result = new PingResult[addresses.length];
                requestsMap = new TreeMap<Integer, Integer>();
                for (int i = 0; i < addresses.length; i++) {
                    int messageId;

                    synchronized (requestsMap) {
                        if (dataToSend != null)
                            messageId = Ping.echo(addresses[i], dataToSend, timeout, this);
                        else
                            messageId = Ping.echo(addresses[i], dataLength, timeout, this);
                        requestsMap.put(messageId, i);
                    }
                }
            } catch (PingErrorException e) {
                throw new PingErrorException("Echo request failure", e);
            }
            countDown.await();
        } else {
            throw new PingErrorException("Illegal parameter", null);
        }
        return result;
    }
}
