/*
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 *
 */
package com.nextian.icmpping.helpers.handlers;

import com.nextian.icmpping.helpers.PingResponseHandler;
import com.nextian.icmpping.jni.IcmpSocket;
import com.nextian.icmpping.util.TimeUtil;

import java.net.InetAddress;


/**
 * Class contains data for passing ICMP request between application, ICMP {@link Sender} and
 * ICMP {@link Receiver}.
 */
public class PingRequest implements Comparable<PingRequest> {
    /**
     * Default request data when data==null is passed to PingRequest constructor.
     */
    static private final byte[] DEFAULT_DATA = new byte[IcmpSocket.MAX_ICMP_DATA_SIZE];

    /**
     * Pattern for filling {@link #DEFAULT_DATA}.
     */
    static private final byte[] DEFAULT_DATA_PATTERN = {0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77};

    static {
        // Initialize DEFAULT_DATA
        for (int dest = 0; dest < DEFAULT_DATA.length; dest += DEFAULT_DATA_PATTERN.length) {
            System.arraycopy(DEFAULT_DATA_PATTERN, 0, DEFAULT_DATA, dest, Math.min(DEFAULT_DATA_PATTERN.length, DEFAULT_DATA.length - dest));
        }
    }

    /**
     * Unique ICMP echo request message identifier.
     */
    private final int messageId;

    /**
     * Handler executed for ICMP response or timeout.
     */
    private final PingResponseHandler handler;

    /**
     * Request destination address.
     */
    private final InetAddress address;

    /**
     * Data sent in the request.
     */
    private final byte[] data;

    /**
     * Length of {@link #data}.
     */
    private final int dataLength;

    /**
     * Get time when object was created (set to current time in constructor.
     */
    private final long creationTime;

    /**
     * Request timeout in milliseconds.
     */
    private final long timeout;

    /**
     * Time the request was handled by sender (set to current time in constructor, later
     * adjusted by {@link #updateSendTime()}.
     */
    private long sendTime;

    /**
     * Initializing constructor. The only attribute changed after the constructor is {@link #sendTime}
     * (it is changed when object is passed to {@link Sender}).
     *
     * @param address    target host address
     * @param data       data to be sent (when null {@link #DEFAULT_DATA} is used)
     * @param dataLength length of data to be sent (cannot be bigger than data buffer length)
     * @param timeout    send timeout in milliseconds
     * @param messageId  ICMP message id
     * @param handler    request handler object
     */
    public PingRequest(InetAddress address, byte[] data, int dataLength, long timeout, int messageId, PingResponseHandler handler) {
        creationTime = TimeUtil.currentTimeMillis();
        this.address = address;
        if (data != null) {
            this.data = data;
        } else {
            this.data = DEFAULT_DATA;
        }
        this.dataLength = dataLength;
        this.timeout = timeout;
        this.messageId = messageId;
        this.handler = handler;
        sendTime = 0;
    }

    /**
     * Copy constructor.
     *
     * @param pingRequest source object to be copied
     */
    public PingRequest(PingRequest pingRequest) {
        creationTime = pingRequest.getCreationTime();
        address = pingRequest.getAddress();
        data = pingRequest.getData();
        dataLength = pingRequest.getDataLength();
        messageId = pingRequest.getMessageId();
        timeout = pingRequest.getTimeout();
        handler = pingRequest.getHandler();
        sendTime = pingRequest.getSendTime();
    }

    public int getMessageId() {
        return messageId;
    }

    public PingResponseHandler getHandler() {
        return handler;
    }

    public InetAddress getAddress() {
        return address;
    }

    public byte[] getData() {
        return data;
    }

    public int getDataLength() {
        return dataLength;
    }

    public long getCreationTime() {
        return creationTime;
    }

    public long getTimeout() {
        return timeout;
    }

    public long getSendTime() {
        return sendTime;
    }

    /**
     * Get request wait end time (creationTime + timeout).
     *
     * @return request end time in milliseconds
     */
    public long getWaitEndTime() {
        return creationTime + timeout;
    }

    /**
     * Set {@link #sendTime} to when the request was actually sent by the sender.
     */
    public void updateSendTime() {
        sendTime = TimeUtil.currentTimeMillis();
    }

    /**
     * Compares this to another request based on values of wait end times returned by {@link #getWaitEndTime()}.
     *
     * @param pingRequest object to compare this against
     * @return 1, 0 or -1
     */
    public int compareTo(PingRequest pingRequest) {
        return Long.signum(getWaitEndTime() - pingRequest.getWaitEndTime());
    }
}
