/*
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 *
 */
package com.nextian.icmpping.helpers.handlers;

import com.nextian.icmpping.IcmpStatus;
import com.nextian.icmpping.helpers.exceptions.PingErrorException;
import com.nextian.icmpping.jni.IcmpEchoResponse;
import com.nextian.icmpping.jni.IcmpSocket;
import com.nextian.icmpping.jni.exception.IcmpException;
import com.nextian.icmpping.jni.exception.IcmpTimeoutException;
import com.nextian.icmpping.util.TimeUtil;

import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Receives ICMP responses and processes them.
 */
public class Receiver implements Runnable {
    private final IcmpSocket icmp;
    private final Lock mutex;
    private final Condition hasData;
    private Queue<PingRequest> requestsQueue;

    /**
     * Initializing constructor.
     *
     * @param icmpSocket socket for receiving responses
     */
    public Receiver(IcmpSocket icmpSocket) {
        this.icmp = icmpSocket;
        requestsQueue = new PriorityQueue<PingRequest>();
        mutex = new ReentrantLock();
        hasData = mutex.newCondition();
    }

    /*
     * Validate ICMP response.
     * Checks if ICMP response matches request parameters. It is
     * needed because ICMP socket receives all ICMP messages accepted by network interfaces.
     */
    private boolean isValidResponse(PingRequest request, IcmpEchoResponse responseMsg) {
        boolean result;

        result = (responseMsg.getStatus() != IcmpStatus.UnexpectedResponse)
                && (responseMsg.getMessageId() == request.getMessageId())
                && (responseMsg.getDestinationAddress().equals(request.getAddress()));
        return result;
    }

    /**
     * Register echo request and put it into requests queue.
     *
     * @param request request to register
     * @throws PingErrorException exceptions is thrown if <code>Receiver</code> object is disposed
     */
    public void registerRequest(PingRequest request) throws PingErrorException {
        mutex.lock();
        try {
            if (requestsQueue != null) {
                PingRequest requestCopy = new PingRequest(request);
                requestsQueue.add(requestCopy);
                hasData.signal();
            } else {
                throw new PingErrorException("Not initialized", null);
            }
        } finally {
            mutex.unlock();
        }
    }

    @Override
    public void run() {
        IcmpEchoResponse responseMsg;
        int timeoutTime;
        long endTime;

        try {
            while (!Thread.interrupted()) {
                try {
                    // Take smallest end time value to handle timeout
                    endTime = getMinimumEndTime();
                    timeoutTime = (int) (endTime - TimeUtil.currentTimeMillis());

                    // Wait for message or handle timeout (via exceptions)
                    icmp.select(timeoutTime);

                    // Receive ICMP message
                    responseMsg = icmp.receive();

                    // Handle received response
                    handleResponse(responseMsg);
                } catch (IcmpTimeoutException e) {
                    handleTimeout();
                }
            }
        } catch (IcmpException ex) {
            cleanupRequestsQueue(ex);
        } catch (InterruptedException ex) {
            cleanupRequestsQueue(ex);
        }
    }

    /*
     * Handle timeout condition.
     */
    private void handleTimeout() {
        PingRequest request;

        mutex.lock();
        try {
            request = requestsQueue.poll();
            request.getHandler().handleTimeout(request);
        } finally {
            mutex.unlock();
        }
    }

    /*
     * Handle received response.
     */
    private void handleResponse(IcmpEchoResponse responseMsg) {
        PingRequest request;
        Iterator<PingRequest> it;

        mutex.lock();
        try {
            it = requestsQueue.iterator();
            while (it.hasNext()) {
                request = it.next();
                if (isValidResponse(request, responseMsg)) {
                    request.getHandler().handleResponse(request, responseMsg);
                    it.remove();
                }
            }
        } finally {
            mutex.unlock();
        }
    }

    /*
     * Empty pending requests queue calling exception handler for each request.
     */
    private void cleanupRequestsQueue(Exception e) {
        PingRequest request;

        mutex.lock();
        try {
            request = requestsQueue.poll();
            while (request != null) {
                request.getHandler().handleException(request, new PingErrorException("ICMP socket fatal error", e));
                request = requestsQueue.poll();
            }
            requestsQueue = null;
        } finally {
            mutex.unlock();
        }
    }

    /*
     * Get end time of request that will be timed out first (it is first one in sorted queue).
     * If there is no request in queue, operation waits until the queue is not empty.
     */
    private long getMinimumEndTime() throws InterruptedException {
        long result;

        mutex.lock();
        try {
            while (requestsQueue.isEmpty()) {
                hasData.await();
            }
            result = requestsQueue.peek().getWaitEndTime();
        } finally {
            mutex.unlock();
        }
        return result;
    }
}
