/*
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 *
 */
package com.nextian.icmpping.helpers.handlers;

import com.nextian.icmpping.helpers.exceptions.PingErrorException;
import com.nextian.icmpping.jni.IcmpSocket;
import com.nextian.icmpping.jni.exception.IcmpSystemException;

import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Ping request sender.
 */
public class Sender implements Runnable {
    private final IcmpSocket icmpSocket;
    private final Queue<PingRequest> requestsQueue;
    private final Lock mutex;
    private final Condition hasData;
    private final Receiver receiver;

    /**
     * Initializing constructor.
     *
     * @param receiver   receiver where sent requests are passed for receiving
     * @param icmpSocket socket used for sending requests
     */
    public Sender(Receiver receiver, IcmpSocket icmpSocket) {
        this.icmpSocket = icmpSocket;
        this.receiver = receiver;
        requestsQueue = new ConcurrentLinkedQueue<PingRequest>();
        mutex = new ReentrantLock();
        hasData = mutex.newCondition();
    }

    /**
     * Enqueue request for processing.
     *
     * @param pingRequest request to process
     * @throws PingErrorException when {@link #requestsQueue} is null (not initialized, or already disposed)
     */
    public void registerRequest(PingRequest pingRequest) throws PingErrorException {
        mutex.lock();
        try {
            if (requestsQueue != null) {
                requestsQueue.add(pingRequest);
                hasData.signal();
            } else {
                throw new PingErrorException("Not initialized:", null);
            }
        } finally {
            mutex.unlock();
        }
    }

    @Override
    public void run() {
        PingRequest request;

        try {
            while (!Thread.interrupted()) {
                // Peek request (wait for request if queue is empty).
                // Request remains in queue until it is successfully registered in receiver.
                request = peekRequest();
                // Register request in receiver so receiver expects its response.
                // If operation succeeds, request is temporary stored in two queues.
                request.updateSendTime();
                receiver.registerRequest(request);
                // After request is registered in receiver, it ma be removed from sender queue.
                // The code between "receiver.registerRequest" and "requestsQueue.poll" is not
                // interruptible, so we can be sure the request is not processed by two times.
                requestsQueue.poll();
                // Send echo.
                icmpSocket.send(request.getAddress(), request.getMessageId(), request.getData(), request.getDataLength());
            }
        } catch (PingErrorException e) {
            cleanupRequestsQueue(e);
        } catch (IcmpSystemException e) {
            cleanupRequestsQueue(e);
        } catch (InterruptedException e) {
            cleanupRequestsQueue(e);
        }
    }

    /*
     * Clean up request queue by calling exception handler for all pending requests.
     */
    private void cleanupRequestsQueue(Exception e) {
        PingRequest request;

        mutex.lock();
        try {
            request = requestsQueue.poll();
            while (request != null) {
                request.getHandler().handleException(request, new PingErrorException("Operation terminated", e));
                request = requestsQueue.poll();
            }
        } finally {
            mutex.unlock();
        }
    }

    /*
     * Get the first pending request (wait if none is immediately available).
     */
    private PingRequest peekRequest() throws InterruptedException {
        PingRequest result;

        mutex.lock();
        try {
            while (requestsQueue.isEmpty()) {
                hasData.await();
            }
            result = requestsQueue.peek();
        } finally {
            mutex.unlock();
        }
        return result;
    }
}
