/*
 *
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.nextian.icmpping.jni;

import com.nextian.icmpping.IcmpStatus;
import com.nextian.icmpping.util.TimeUtil;

import java.net.InetAddress;
import java.util.TreeMap;

/**
 * Simplified ICMP echo response message (only necessary fields are included).
 */
public class IcmpEchoResponse {
    /**
     * Special datagram conversion error value for {@link #type}.
     */
    private final static byte DatagramConversionError = 31;
    /**
     * A map for conversion of integers to {@link IcmpStatus}, used by {@link #icmpStatusByInt}.
     */
    private static final TreeMap<Integer, IcmpStatus> icmpStatusMap;

    static {
        icmpStatusMap = new TreeMap<Integer, IcmpStatus>();
        for (IcmpStatus status : IcmpStatus.values()) {
            icmpStatusMap.put(status.getValue(), status);
        }
    }

    /**
     * ICMP echo request message id combined as integer from identifier+sequenceNumber (see RFC792),
     * used for unique message identification.
     */
    private int messageId;
    /**
     * ICMP message type (see RFC792).
     */
    private byte type;
    /**
     * ICMP message code (see RFC792).
     */
    private byte code;
    /**
     * Data received in the ICMP response.
     */
    private byte[] data;
    /**
     * Source address from which this ICMP request is originating.
     */
    private InetAddress sourceAddress;
    /**
     * Destination address for this ICMP request.
     */
    private InetAddress destinationAddress;
    /**
     * Object creation time in milliseconds, used as response time.
     */
    private long creationTime;

    /**
     * The default and only constructor, sets {@link #creationTime}. This value has no other use
     * then enabling time difference measurements in milliseconds using {@link TimeUtil#currentTimeMillis()}.
     */
    public IcmpEchoResponse() {
        creationTime = TimeUtil.currentTimeMillis();
    }

    /**
     * Convert integer value to corresponding {@link IcmpStatus} enum.
     *
     * @param value input integer value to convert
     * @return corresponding IcmpStatus, or IcmpStatus.UnexpectedResponse when code cannot be found
     */
    private static IcmpStatus icmpStatusByInt(int value) {
        IcmpStatus result = icmpStatusMap.get(value);
        if (result == null) {
            result = IcmpStatus.UnexpectedResponse;
        }
        return result;
    }

    /*
     * Accessor to {@link #code} (ICMP message code, see RFC792).
     */
    private byte getCode() {
        return code;
    }

    /**
     * Accessor to {@link #data } (data received in ICMP response).
     * This value is populated only when <code>EchoReply</code> message is received.
     *
     * @return data received
     */
    public byte[] getData() {
        return data;
    }

    /**
     * Accessor to #messageId (combined as integer from identifier+sequenceNumber, see RFC792).
     *
     * @return identifier of ICMP message
     */
    public int getMessageId() {
        return messageId;
    }

    /**
     * Accessor to {@link #type} (ICMP message type, see RFC792).
     */
    private byte getType() {
        return type;
    }

    /**
     * Retrieve address of a source host for the ICMP response. On success, this value is set to address of the host
     * provided as destination in the corresponding ICMP request. In case network rejected the request,
     * it is set to address of the host that rejected the message.
     *
     * @return network address of host providing the response
     */
    public InetAddress getSourceAddress() {
        return sourceAddress;
    }

    /**
     * Get address of host the ICMP request was sent to.
     *
     * @return network address of destination host
     */
    public InetAddress getDestinationAddress() {
        return destinationAddress;
    }

    /**
     * Get status of ICMP response. The result enumeration describes type of ICMP response and is calculated
     * from <i>type</i> and <i>code</i> attributes of ICMP response message as defined in RFC 792.
     *
     * @return status if ICMP processing
     */
    public IcmpStatus getStatus() {
        IcmpStatus result;

        // Class reports ICMP type 31 (Datagram Conversion ERROR) without detailed error code so no code is added
        // to status value
        if (getType() != DatagramConversionError) {
            result = icmpStatusByInt(getType() * 256 + getCode());
        } else {
            result = icmpStatusByInt(getType() * 256);
        }
        return result;
    }

    /**
     * RESPONSE timestamp is equal to the moment that the object was created.
     *
     * @return time when response was received in milliseconds
     */
    public long getResponseTime() {
        return creationTime;
    }
}
