/*
 *
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.nextian.icmpping.jni;

import com.nextian.icmpping.jni.exception.IcmpFatalErrorException;
import com.nextian.icmpping.jni.exception.IcmpSystemException;
import com.nextian.icmpping.jni.exception.IcmpTimeoutException;

import java.net.InetAddress;

/**
 * W wrapper class encapsulating raw sockets and methods from icmp-helpers.dll.
 * Each instance of IcmpSocket object has its own ICMP socket handle.
 * Underneath, a raw socket is created, requiring <i>Administrator</i> privileges.
 * Including this class in your application causes <i>icmp-helpers.dll</i> be loaded, the dll
 * must be available in one of the directories specified by <i>java.library.path</i>.
 * The IcmpSocket object is allocating system resources when successful call to {@link #open}
 * is performed. In order to release those resources the {@link #close} method must be executed.
 */
public class IcmpSocket {
    /**
     * Maximum ICMP data size.
     */
    public static final int MAX_ICMP_DATA_SIZE = 65535 - 28;
    /**
     * DLL name (no extension suffix or architecture string).
     */
    private static final String DLL_NAME = "icmp-ping";
    /**
     * Windows/WinSock socket error return code.
     */
    private static final long SOCKET_ERROR = -1;
    /**
     * Windows/WinSock invalid socket handle.
     */
    private static final long INVALID_SOCKET_HANDLE = -1;

    /*
     * Load DLL module. The actual native library is icmp-helpers-&lt;os.arch;&gt.
     *
     */
    static {
        final String libName = DLL_NAME + "-" + System.getProperty("os.arch");
        System.loadLibrary(libName);
    }

    /**
     * Socket handle encapsulated by the class.
     */
    private long handle = INVALID_SOCKET_HANDLE;

    /**
     * If socket is still open (was not explicitly closed), close it.
     */
    @Override
    protected void finalize() throws Throwable {
        if (handle != INVALID_SOCKET_HANDLE) {
            close();
        }
        super.finalize();
    }

    /**
     * Obtain a guaranteed, system-wide unique ICMP message id in the range of 0 up to 2147483647.
     * The identifier value is increased on every call. When maximum value is reached
     * counting is restarted from 0.
     * The counter value is common for all applications using icmp-helpers.dll module
     * so they never get duplicate value even if DLL is loaded by multiple JVMs.
     *
     * @return unique integer ICMP message id
     */
    public int getNextMessageId() {
        return nativeGetNextMessageId();
    }

    /**
     * Open raw ICMP socket. If socket is already opened, no action is performed.
     * To avoid resource leaks the socket must be closed by calling {@link #close()}
     *
     * @throws IcmpSystemException on open error
     */
    public synchronized void open() throws IcmpSystemException {
        if (handle == INVALID_SOCKET_HANDLE) {
            handle = nativeCreateSocket();
        }
    }

    /**
     * Close ICMP socket opened with {@link #open()}. If socket has not been opened, no action is performed.
     *
     * @throws IcmpSystemException if socket cannot be closed
     */
    public synchronized void close() throws IcmpSystemException {
        if (handle != INVALID_SOCKET_HANDLE) {
            nativeCloseSocket(handle);
            handle = INVALID_SOCKET_HANDLE;
        }
    }

    /**
     * Wait for ICMP socket data.
     * Method waits until data is available for reading or given timeout is reached.
     * Closing socket ({@link #close}) terminates operation and causes method to throw exceptions.
     *
     * @param timeout time to wait for data in milliseconds
     * @throws IcmpSystemException  underlying WinSock call fails
     * @throws IcmpTimeoutException timeout occurred
     */
    public void select(int timeout) throws IcmpSystemException, IcmpTimeoutException {
        nativeSelectSocket(handle, timeout);
    }

    /**
     * Send ICMP echo request (ICMP socket must already be opened using {@link #open}).
     * The <code>messageId</code> is split into ICMP message identifier and
     * sequenceNumber (see RFC 792). The <code>data</code> is sent as ICMP optional data field.
     * The <code>messageId</code> should be unique value because it is used to recognize
     * response message.
     *
     * @param address    target host address
     * @param messageId  unique identifier of request message
     * @param data       data to be sent
     * @param dataLength length of data to be send
     * @return number of bytes actually sent
     * @throws IcmpSystemException underlying WinSock call fails
     */
    public int send(InetAddress address, int messageId, byte[] data, int dataLength) throws IcmpSystemException {
        int result;

        result = nativeSendEcho(handle, address.getAddress(), messageId, data, dataLength);
        return result;
    }

    /**
     * Receive ICMP response.
     * Before this method is used the ICMP socket must be created using {@link #open}.
     * The received response is mapped to {@link IcmpEchoResponse} object. The response object
     * content depends an type of received ICMP response, thus, before accessing <code>IcmpEchoResponse</code>
     * values the status of response should be verified using {@link IcmpEchoResponse#getStatus} method.
     * This is a <b>blocking</method>. Use {@link #select} to detect when data is available.
     *
     * @return received ICMP response message (it's never null, as in case any problem an exceptions is thrown)
     * @throws IcmpSystemException     underlying WinSock call fails
     * @throws IcmpFatalErrorException JNI problem
     */
    public IcmpEchoResponse receive() throws IcmpSystemException, IcmpFatalErrorException {
        IcmpEchoResponse result;

        result = nativeReceiveMessage(handle);
        return result;
    }

    // Native methods from the DLL:
    private native int nativeSendEcho(long handle, byte[] destAddress, int messageId, byte[] dataBuffer, int dataLength) throws IcmpSystemException;

    private native IcmpEchoResponse nativeReceiveMessage(long handle) throws IcmpSystemException, IcmpFatalErrorException;

    private native long nativeCreateSocket() throws IcmpSystemException;

    private native void nativeCloseSocket(long handle) throws IcmpSystemException;

    private native void nativeSelectSocket(long handle, int timeout) throws IcmpSystemException, IcmpTimeoutException;

    private native int nativeGetNextMessageId();
}
