/*
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.nextian.icmpping;

import com.nextian.icmpping.helpers.Receiver;
import com.nextian.icmpping.helpers.Sender;
import com.nextian.icmpping.jni.IcmpEchoResponse;
import com.nextian.icmpping.jni.IcmpSocket;
import com.nextian.icmpping.jni.exception.IcmpFatalErrorException;
import com.nextian.icmpping.jni.exception.IcmpSystemException;
import com.nextian.icmpping.jni.exception.IcmpTimeoutException;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import org.apache.log4j.Logger;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * ICMP JNI interface unit test.
 */
public class IcmpSocketTest extends TestCase {

    private static final Logger LOGGER = Logger.getLogger(IcmpSocketTest.class);

    /**
     * Create the test case.
     *
     * @param testName test case name
     */
    public IcmpSocketTest(String testName) {
        super(testName);
    }

    /**
     * Build test suite.
     *
     * @return the suite of tests being tested
     */
    public static Test suite() {
        return new TestSuite(IcmpSocketTest.class);
    }

    /**
     * Test create ICMP raw socket.
     */
    public void testCreateIcmpJniBridge() {
        IcmpSocket icmpSocket;
        try {
            icmpSocket = new IcmpSocket();
            assertNotNull(icmpSocket);
            icmpSocket.open();
            icmpSocket.close();
        } catch (IcmpSystemException e) {
            fail(e.getMessage());
        }
    }

    /**
     * Send ICMP echo message is sent - success is when a response is received. Google Public DNS address 8.8.8.8
     * is used as it should be always responding.
     */
    public void testBasicSendReceive() {
        final byte[] data = "testBasicSendReceive".getBytes();
        final int waitTimeout = 10000; // 10 seconds
        IcmpSocket icmpSocket;
        IcmpEchoResponse echoResponse;

        try {
            icmpSocket = new IcmpSocket();
            Inet4Address address = (Inet4Address) InetAddress.getByName("8.8.8.8");
            icmpSocket.open();
            icmpSocket.send(address, 1, data, data.length);
            icmpSocket.select(waitTimeout);
            echoResponse = icmpSocket.receive();
            LOGGER.debug(echoResponse.toString());
            icmpSocket.close();
        } catch (IcmpFatalErrorException e) {
            fail(e.getMessage());
        } catch (IcmpTimeoutException e) {
            fail(e.getMessage());
        } catch (UnknownHostException e) {
            fail(e.getMessage());
        } catch (IcmpSystemException e) {
            fail(e.getMessage());
        }
    }

    /**
     * Test ICMP JNI interface by sending and receiving data on single IcmpSocket
     * object using two threads (sending and receiving threads). Again, no API
     * errors are allowed.
     */
    public void testMultiThreadSendReceive() {
        final int requestsToSend = 4000;
        final int waitTimeout = 5000;

        AtomicInteger sentCounter;
        AtomicInteger responseCount;
        IcmpSocket icmpSocket;
        Sender sendHandler;
        Receiver receiveHandler;
        Thread sendThread;
        Thread receiveThread;

        try {
            sentCounter = new AtomicInteger();
            responseCount = new AtomicInteger();
            icmpSocket = new IcmpSocket();
            icmpSocket.open();

            sendHandler = new Sender(LOGGER, icmpSocket, requestsToSend, sentCounter);
            receiveHandler = new Receiver(LOGGER, icmpSocket, responseCount, waitTimeout);

            sendThread = new Thread(sendHandler);
            receiveThread = new Thread(receiveHandler);

            receiveThread.start();
            sendThread.start();

            sendThread.join();
            receiveThread.join();

            icmpSocket.close();

            LOGGER.info("Requests sent: " + sentCounter.get());
            LOGGER.info("Responses received: " + responseCount.get());
        } catch (InterruptedException e) {
            fail(e.getMessage());
        } catch (IcmpSystemException e) {
            fail(e.getMessage());
        }
    }
}
