/*
 *
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.nextian.icmpping;

import com.nextian.icmpping.helpers.Ping;
import com.nextian.icmpping.helpers.ResponseHandler;
import com.nextian.icmpping.helpers.TestData;
import com.nextian.icmpping.helpers.exceptions.PingErrorException;
import org.apache.log4j.Logger;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

/**
 * Long running helpers test. Each iteration, memory usage is reported enabling memory leak checking.
 * This is <b>not a junit test</b>.
 */
public class LongRunningPingTest {
    /**
     * Class logger.
     */
    private static final Logger LOGGER = Logger.getLogger(LongRunningPingTest.class);

    /**
     * Number of iterations (each iteration sends {@link #REQUEST_COUNT} helpers requests.
     */
    private static final int ITERATIONS_COUNT = 10;
    /**
     * Number of requests sent per each iteration.
     */
    private static final int REQUEST_COUNT = 1000;
    /**
     * TIMEOUT in milliseconds for each request.
     */
    private static final int PING_TIMEOUT = 2000; // 2 seconds

    /**
     * Common test data for all the tests (initialized in constructor)
     */
    private final TestData testData;

    /**
     * Constructor (initialize test data)
     */
    private LongRunningPingTest() {
        testData = new TestData();
    }

    /**
     * Test Ping echo requests using predefined set of hosts.
     */
    public static void main(String[] args) {
        LongRunningPingTest longRunningPingTest = new LongRunningPingTest();
        longRunningPingTest.run();
    }

    /**
     * Run all test iterations (main test entry point).
     */
    private void run() {
        for (int i = 0; i < ITERATIONS_COUNT; i++) {
            LOGGER.info("Iteration " + i + " of " + ITERATIONS_COUNT);
            runIteration();
        }
    }

    /**
     * Run a single test iteration.
     */
    private void runIteration() {
        CountDownLatch countDownLatch = new CountDownLatch(REQUEST_COUNT);
        ResponseHandler handler = new ResponseHandler(countDownLatch, LOGGER);

        try {
            Ping.init();
            for (int j = 0; j < REQUEST_COUNT; j++) {
                Ping.echo(testData.getNextHost(), TestData.DEFAULT_ICMP_TEST_PACKET_DATA, PING_TIMEOUT, handler);
            }
            // Success when countDownLatch reaches 0 before timeout
            countDownLatch.await(PING_TIMEOUT + 100 /* add extra 100ms */, TimeUnit.MILLISECONDS);
            LOGGER.info(
                    new StringBuilder()
                            .append("Responses received:")
                            .append(handler.getResponseCount())
                            .append(" Timeout:")
                            .append(handler.getTimeoutCount())
                            .append(" Errors:")
                            .append(handler.getExceptionCount())
                            .toString());
        } catch (PingErrorException ex) {
            LOGGER.error("Test initialization error: " + ex);
        } catch (InterruptedException ex) {
            LOGGER.error("Test interrupted");
        } finally {
            Ping.release();
            LOGGER.info(
                    new StringBuilder()
                            .append("Memory: free: ")
                            .append(Runtime.getRuntime().freeMemory())
                            .append(" total: ")
                            .append(Runtime.getRuntime().totalMemory())
                            .append(" max: ")
                            .append(Runtime.getRuntime().maxMemory())
                            .toString());
        }
    }
}
