/*
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 *
 */

package com.nextian.icmpping;

import com.nextian.icmpping.helpers.Ping;
import com.nextian.icmpping.helpers.PingResult;
import com.nextian.icmpping.helpers.ResponseHandler;
import com.nextian.icmpping.helpers.TestData;
import com.nextian.icmpping.helpers.exceptions.PingErrorException;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import org.apache.log4j.Logger;

import java.net.InetAddress;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

/**
 * Unit tests for {@link Ping}.
 */
public class PingTest extends TestCase {
    private static final Logger LOGGER = Logger.getLogger(PingTest.class);

    private final TestData testData;

    /**
     * Create test case: call super constructor and initialize test data.
     *
     * @param name test case name
     */
    public PingTest(String name) {
        super(name);
        testData = new TestData();
    }

    /**
     * Create test suite.
     *
     * @return suite of tests for the class
     */
    public static Test suite() {
        return new TestSuite(PingTest.class);
    }

    /**
     * Test initialization/shutdown.
     */
    public void testPingInit() {
        try {
            Ping.init();
            Ping.release();
        } catch (PingErrorException e) {
            fail(e.getMessage());
        }
    }

    /**
     * Asynchronous ping test with handler.
     */
    public void testAsynchronousPingWithHandler() {
        final int pingTimeout = 1000;
        final int waitTimeout = pingTimeout * 2; // this should be sufficient for all pings to terminate
        CountDownLatch counter;
        ResponseHandler handler;

        try {
            counter = new CountDownLatch(testData.getDestinationAddresses().length);
            handler = new ResponseHandler(counter, LOGGER);

            Ping.init();
            for (InetAddress destinationAddress : testData.getDestinationAddresses()) {
                Ping.echo(destinationAddress, TestData.DEFAULT_DATA_LENGTH, pingTimeout, handler);
            }
            assertTrue(counter.await(waitTimeout, TimeUnit.MILLISECONDS));
            assertTrue(handler.getTotalResultCount() == testData.getDestinationAddresses().length);
        } catch (PingErrorException e) {
            fail(e.getMessage());
        } catch (InterruptedException e) {
            fail("Test interrupted");
        } finally {
            Ping.release();
        }
    }

    /**
     * Test blocking ping passing specific data for the ICMP packet.
     */
    public void testBlockingPingWithSpecificData() {
        final int pingTimeout = 5000;
        PingResult[] pingResults;

        try {
            Ping.init();
            pingResults = Ping.echo(testData.getDestinationAddresses(), TestData.DEFAULT_ICMP_TEST_PACKET_DATA, pingTimeout);
            for (PingResult r : pingResults) {
                logPingResult(r);
            }
        } catch (PingErrorException e) {
            fail("Unexpected PingErrorException " + e.getMessage());
        } catch (InterruptedException e) {
            fail("Unexpected InterruptedException" + e.getMessage());
        } finally {
            Ping.release();
        }
    }

    /**
     * Test blocking ping with default generated data for the ICMP packet.
     */
    public void testBlockingPingWithDefaultData() {
        final int pingTimeout = 5000;  // wait timeout for helpers requests
        PingResult[] pingResults;

        try {
            Ping.init();
            pingResults = Ping.echo(testData.getDestinationAddresses(), TestData.DEFAULT_DATA_LENGTH, pingTimeout);
            for (PingResult r : pingResults) {
                logPingResult(r);
            }
        } catch (PingErrorException e) {
            fail("Unexpected PingErrorException " + e.getMessage());
        } catch (InterruptedException e) {
            fail("Unexpected InterruptedException" + e.getMessage());
        } finally {
            Ping.release();
        }
    }

    /**
     * Test interrupting of blocking helpers. On fast machines and networks WAIT_TIMEOUT may have to be decreased
     * for the test complete successfully.
     */
    public void testBlockingPingInterrupt() {
        final int pingTimeout = 5000;  // wait timeout for helpers requests
        final int waitTimeout = 30;    // timeout after which this thread is interrupted
        PingResult[] pingResults;

        try {
            Ping.init();

            LOGGER.debug("Setting up interrupt action");
            new Thread(new InterruptThreadAction(Thread.currentThread(), waitTimeout)).start();

            LOGGER.debug("Issuing blocking requests");
            pingResults = Ping.echo(testData.getDestinationAddresses(), TestData.DEFAULT_ICMP_TEST_PACKET_DATA, pingTimeout);

            fail("Blocking echo completed without interruption");
            for (PingResult r : pingResults) {
                logPingResult(r);
            }
        } catch (PingErrorException e) {
            fail("Unexpected PingErrorException " + e.getMessage());
        } catch (InterruptedException e) {
            LOGGER.debug("Blocking helpers successfully interrupted");
        } finally {
            Ping.release();
        }
    }

    private void logPingResult(PingResult result) {
        LOGGER.debug(
                new StringBuilder()
                        .append("Ping:")
                        .append(result.getAddress().getHostAddress())
                        .append("  time:")
                        .append(result.getResponseTime())
                        .append("  result:")
                        .append(result.getResult())
                        .append("  status:")
                        .append(result.getStatus())
                        .toString());
    }

    private class InterruptThreadAction implements Runnable {
        private final Thread threadToInterrupt;
        private final int timeout;

        InterruptThreadAction(Thread threadToInterrupt, int timeout) {
            this.threadToInterrupt = threadToInterrupt;
            this.timeout = timeout;
        }

        @Override
        public void run() {
            try {
                Thread.sleep(timeout);
            } catch (InterruptedException e) {
                // Ignore the exceptions
            }
            threadToInterrupt.interrupt();
        }
    }
}
