/*
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 *
 */
package com.nextian.icmpping.helpers;

import com.nextian.icmpping.jni.IcmpEchoResponse;
import com.nextian.icmpping.jni.IcmpSocket;
import com.nextian.icmpping.jni.exception.IcmpFatalErrorException;
import com.nextian.icmpping.jni.exception.IcmpSystemException;
import com.nextian.icmpping.jni.exception.IcmpTimeoutException;
import org.apache.log4j.Logger;

import java.util.concurrent.atomic.AtomicInteger;

/**
 * Internally used class for testing socket simultaneous send/receive
 * operation. This handler is responsible for receiving messages from provided
 * IcmpSocket object.
 */
public class Receiver implements Runnable {
    private final Logger logger;
    private final IcmpSocket icmp;
    private final AtomicInteger responseCount;
    private final int timeout;

    /**
     * Create and initialize members.
     *
     * @param logger        error logger
     * @param icmp          the ICMP object messages are received from
     * @param responseCount number of responses before final timeout has been received
     * @param timeout       socket wait timeout
     */
    public Receiver(Logger logger, IcmpSocket icmp, AtomicInteger responseCount, int timeout) {
        this.logger = logger;
        this.icmp = icmp;
        this.responseCount = responseCount;
        this.timeout = timeout;
    }

    public void run() {
        IcmpEchoResponse responseMsg;

        try {
            // Loop until all responses are received or timeout occurs (the timeout is large
            // enough for responses to be received on average):
            while (!Thread.interrupted()) {
                icmp.select(timeout);
                responseMsg = icmp.receive();
                responseCount.incrementAndGet();
            }
        } catch (IcmpTimeoutException e) {
            logger.debug("Final timeout received, exiting Receiver");
        } catch (IcmpFatalErrorException e) {
            logger.error(e.getMessage());
        } catch (IcmpSystemException e) {
            logger.error(e.getMessage());
        }
    }
}
