/*
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 *
 */

package com.nextian.icmpping.helpers;

import com.nextian.icmpping.helpers.handlers.PingRequest;
import com.nextian.icmpping.jni.IcmpEchoResponse;
import com.nextian.icmpping.util.TimeUtil;
import org.apache.log4j.Logger;

import java.util.concurrent.CountDownLatch;

/**
 * Common helpers response handler implementation for tests.
 */
public class ResponseHandler implements PingResponseHandler {
    /**
     * Logger where diagnostic messages are written (passed to and initialized in the constructor).
     */
    private final Logger logger;

    /**
     * Waiting latch passed in constructor used for waiting. It is already set to number of requests we're waiting
     * for and is decremented in every callback.
     */
    private final CountDownLatch countDownLatch;

    /**
     * Number of responses received.
     */
    private int responseCount;

    /**
     * Number of responses that timed out.
     */
    private int timeoutCount;

    /**
     * Number of exceptions.
     */
    private int exceptionCount;

    /**
     * Construct the handler, initialize variables.
     *
     * @param countDownLatch a latch decremented at each callback used for waiting (if all requests completed
     *                       it will be set to zero)
     * @param logger         a logger where messages are written
     */
    public ResponseHandler(CountDownLatch countDownLatch, Logger logger) {
        this.countDownLatch = countDownLatch;
        this.logger = logger;
        responseCount = 0;
        timeoutCount = 0;
        exceptionCount = 0;
    }

    public int getResponseCount() {
        return responseCount;
    }

    public int getTimeoutCount() {
        return timeoutCount;
    }

    public int getExceptionCount() {
        return exceptionCount;
    }

    /**
     * Retrieve sum of all results (responses received, timeouts, exceptions).
     *
     * @return a sum of {@link #responseCount}, {@link #timeoutCount} and {@link #exceptionCount}
     */
    public int getTotalResultCount() {
        return responseCount + timeoutCount + exceptionCount;
    }

    public void handleResponse(PingRequest request, IcmpEchoResponse response) {
        logger.debug(
                new StringBuilder()
                        .append("RESPONSE received for ")
                        .append(request.getAddress().getHostAddress())
                        .append(" from ")
                        .append(response.getSourceAddress().getHostAddress())
                        .append(" Status: ")
                        .append(response.getStatus())
                        .append(" id: ")
                        .append(request.getMessageId())
                        .append(" time:")
                        .append(TimeUtil.currentTimeMillis() - request.getSendTime())
                        .toString());
        countDownLatch.countDown();
        responseCount++;
    }

    public void handleTimeout(PingRequest request) {
        logger.debug(
                new StringBuilder()
                        .append("TIMEOUT for ")
                        .append(request.getAddress().getHostAddress())
                        .append(" id: ")
                        .append(request.getMessageId())
                        .toString());
        countDownLatch.countDown();
        timeoutCount++;
    }

    public void handleException(PingRequest request, Exception exception) {
        logger.debug(
                new StringBuilder()
                        .append("Exception for ")
                        .append(request.getAddress().getHostAddress())
                        .append(" id: ")
                        .append(request.getMessageId())
                        .append(" ")
                        .append(exception.getMessage())
                        .toString());
        countDownLatch.countDown();
        exceptionCount++;
    }
}
