/*
 * Copyright (c) Nextian. All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 *
 */

package com.nextian.icmpping.helpers;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.LinkedList;
import java.util.List;

/**
 * A wrapper around test data for all the tests.
 */
public class TestData {

    /**``
     */
    public static final byte[] DEFAULT_ICMP_TEST_PACKET_DATA = "PingTest: DATA".getBytes();
    /**
     * Default length of helpers data.
     */
    public static final int DEFAULT_DATA_LENGTH = 128;

    /**
     * Target hosts to be pinged (resolved to InetAddresses via {@link #resolveTargetAddresses()}.
     */
    private static final String TEST_HOSTS[] = {
            "127.0.0.1",
            "8.8.8.8",
            "1.2.3.4",
            "news.yahoo.com",
            "www.google.com",
            "www.cnn.com",
            "www.foxnews.com",
            "www.accuweather.com",
            "www.huffingtonpost.com",
            "en.wikipedia.org",
            "news.bbc.co.uk",
            "www.usatoday.com",
            "www.msnbc.com",
            "www.ikea.com",
            "nextian.com",
            "forbes.com"
    };

    /**
     * Destination IP addresses resolved from {@link #TEST_HOSTS}.
     */
    private final InetAddress[] destinationAddresses;
    /**
     * Host index pointing to {@link #destinationAddresses}, for iterating in round robin fashion
     * through the host list {@link #getNextHostIndex()}.
     */
    private int hostIndex = 0;

    public TestData() {
        destinationAddresses = resolveTargetAddresses();
    }

    public InetAddress[] getDestinationAddresses() {
        return destinationAddresses;
    }

    /**
     * Returns subsequent host indexes from 0 to <i>destinationAddresses.length</i> using modulo
     * to wrap. This method is used to pick addresses in round robin from <i>destinationAddresses.length</i>.
     *
     * @return next host index in {@link #destinationAddresses}
     */
    private synchronized int getNextHostIndex() {
        int result = hostIndex;
        hostIndex = (hostIndex + 1) % destinationAddresses.length;
        return result;
    }

    public synchronized InetAddress getNextHost() {
        return destinationAddresses[getNextHostIndex()];
    }

    /**
     * Resolve DNS addresses in {@link #TEST_HOSTS} to an array of InetAddresses.
     *
     * @return array of resolved InetAddress objects (addresses that cannot be resolved are skipped)
     */
    private InetAddress[] resolveTargetAddresses() {
        List<InetAddress> destinationAddresses = new LinkedList<InetAddress>();

        for (String testHost : TEST_HOSTS) {
            InetAddress address;
            try {
                address = Inet4Address.getByName(testHost);
                if (address != null) {
                    destinationAddresses.add(address);
                } else {
                    System.err.println("Unknown host: " + testHost + " - skipping");
                }
            } catch (UnknownHostException ex) {
                System.err.println("Unknown host: " + testHost + " - skipping");
            }
        }
        return destinationAddresses.toArray(new InetAddress[0]);
    }
}
