/*
Copyright (c) Nextian. All rights reserved.

This software is furnished under a license. Use, duplication,
disclosure and all other uses are restricted to the rights
specified in the written license agreement.
*/

#include <wtypes.h>
#include <winbase.h>
#include <limits.h>
#include <tchar.h>
#include "IcmpSocket.h"
#include "Icmp.h"
#include "IcmpDebug.h"
#include "IcmpMutex.h"

#define JNI_ERR_COULD_NOT_SET_FIELD               "Could not set field vale of ICMP response message"
#define JNI_ERR_COULD_NOT_CREATE_RESPONSE_OBJECT  "Could not create ICMP response object"

/*
 * Globally accessible identifier of message.
 * The variable is accessible from all processes that use this DLL. The defined
 * bellow mutex must be used to synchronize access to this variable.
 */
#pragma data_seg(".shared")
int icmpMessageIdentifier = INT_MAX - 10;
#pragma data_seg()

/*
 * Handle of a mutex used to ensure thread safety of IcmpSocket method calls.
 */
static HANDLE mutex = NULL;

/*
 * Windows specific DLL handler function. It is invoked on each DLL load/unload condition.
 * Here, it is used to create and close mutex object (handle). Because
 * it is handled by system we are sure that each mutex creation will be closed.
 */
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	BOOL result = TRUE;

	ICMPOUTPUTDEBUGSTRING0(_T("IcmpPingJniDll: DllMain IN\n"));
	switch (fdwReason)
	{
	case DLL_PROCESS_ATTACH:
		ICMPOUTPUTDEBUGSTRING0(_T("DLL_PROCESS_ATTACH\n"));
		mutex = IcmpCreateMutex();
		result = mutex != NULL;
		break;
	case DLL_THREAD_ATTACH:
		break;
	case DLL_THREAD_DETACH:
		break;
	case DLL_PROCESS_DETACH:
		result = IcmpCloseMutex(mutex);
		ICMPOUTPUTDEBUGSTRING0(_T("DLL_PROCESS_DETACH\n"));
		break;
	default:
		break;
	}
	ICMPOUTPUTDEBUGSTRING1(_T("IcmpPingJniDll: DllMain OUT: %d\n"), result);
	return result;
	UNREFERENCED_PARAMETER(hinstDLL);
	UNREFERENCED_PARAMETER(lpvReserved);
}

/*
 * On load from Java return version of JNI that should be supported by Java.
 * Initialize IcmpLib module.
 */
JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *vm, void *reserved)
{
	ICMPOUTPUTDEBUGSTRING0(_T("ICMP: load"));
	IcmpInitialize();
	return JNI_VERSION_1_4;
	UNREFERENCED_PARAMETER(reserved);
	UNREFERENCED_PARAMETER(vm);
}

/*
 * Release resources allocated in IcmpLib module
 */
JNIEXPORT void JNICALL JNI_OnUnload(JavaVM *vm, void *reserved)
{
	IcmpRelease();
	ICMPOUTPUTDEBUGSTRING0(_T("ICMP: unload"));
	UNREFERENCED_PARAMETER(reserved);
	UNREFERENCED_PARAMETER(vm);
}

/*
 * Throw "com/nextian/icmp/jni/exception/IcmpTimeoutException".
 * The new exception object is created an passed to Java virtual machine
 */
static void throwTimeoutException(JNIEnv *env)
{
	jclass icmpTimeoutExceptionClass;
	jmethodID icmpTimeoutExceptionConstructorId;
	jobject exceptionObject;

	icmpTimeoutExceptionClass = (*env)->FindClass(env, "com/nextian/icmpping/jni/exception/IcmpTimeoutException");
	if (icmpTimeoutExceptionClass != NULL)
	{
		icmpTimeoutExceptionConstructorId = (*env)->GetMethodID(env, icmpTimeoutExceptionClass, "<init>", "()V");
		if (icmpTimeoutExceptionConstructorId != NULL)
		{
			exceptionObject = (*env)->NewObject(env, icmpTimeoutExceptionClass, icmpTimeoutExceptionConstructorId);
			if (exceptionObject != NULL)
			{
				(*env)->Throw(env, exceptionObject);
			}
			else ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Could not create exception object"));
		}
		else ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Class 'com/nextian/icmpping/jni/exception/IcmpTimeoutException' has no required constructor"));
	}
	else ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: No class 'com/nextian/icmpping/jni/exception/IcmpTimeoutException'"));
}

/*
 * Throw "com/nextian/icmpping/jni/exception/IcmpSystemException".
 * The new exception object is created an passed to Java virtual machine.
 * Based on provided errorCode parameter the error string is obtained
 * and both are passed to exception constructor.
 */
static void throwSystemException(JNIEnv *env, int errorCode)
{
	jstring javaString;
	jclass icmpSystemExceptionClass;
	jmethodID icmpSystemExceptionConstructorId;
	jobject exceptionObject;
	LPTSTR errorStr;

	icmpSystemExceptionClass = (*env)->FindClass(env, "com/nextian/icmpping/jni/exception/IcmpSystemException");
	if (icmpSystemExceptionClass != NULL)
	{
		icmpSystemExceptionConstructorId = (*env)->GetMethodID(env, icmpSystemExceptionClass, "<init>", "(ILjava/lang/String;)V");
		if (icmpSystemExceptionConstructorId != NULL)
		{
			errorStr = IcmpGetErrorString(errorCode);
			ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: Error string '%s'"), errorStr);
#ifdef _UNICODE
			javaString = (*env)->NewString(env, errorStr, (jsize)(_tcslen(errorStr)));
#else
			javaString = (*env)->NewStringUTF(env, errorStr);
#endif
			exceptionObject = (*env)->NewObject(env, icmpSystemExceptionClass, icmpSystemExceptionConstructorId, (jint)errorCode, javaString);
			if (exceptionObject != NULL)
			{
				(*env)->Throw(env, exceptionObject);
			}
			else ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Could not create exception object"));
			IcmpReleaseErrorString(errorStr);
		}
		else ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Class 'com/nextian/icmpping/jni/exception/IcmpSystemException' has no required constructor"));
	}
	else ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: No class 'com/nextian/icmpping/jni/exception/IcmpSystemException'"));
}

/*
 * Throw "com/nextian/icmpping/jni/exception/IcmpFatalErrorException".
 * The new exception object is created an passed to Java virtual machine.
 * Provided errorString is passed to exception constructor.
 */
static void throwFatalErrorEception(JNIEnv *env, const char *errorString)
{
	jclass icmpFatalErrorExceptionClass;
	jmethodID icmpFatalErrorConstructorId;
	jobject exceptionObject;
	jstring javaString;

	icmpFatalErrorExceptionClass = (*env)->FindClass(env, "com/nextian/icmpping/jni/exception/IcmpFatalErrorException");
	if (icmpFatalErrorExceptionClass != NULL)
	{
		icmpFatalErrorConstructorId = (*env)->GetMethodID(env, icmpFatalErrorExceptionClass, "<init>", "(Ljava/lang/String;)V");
		if (icmpFatalErrorConstructorId != NULL)
		{
			javaString = (*env)->NewStringUTF(env, errorString);
			exceptionObject = (*env)->NewObject(env, icmpFatalErrorExceptionClass, icmpFatalErrorConstructorId, javaString);
			if (exceptionObject != NULL)
			{
				(*env)->Throw(env, exceptionObject);
			}
			else
			{
				ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Could not create object for class 'com/nextian/icmpping/jni/exception/IcmpFatalErrorException'"));
			}
		}
		else
		{
			ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Class 'com/nextian/icmpping/jni/exception/IcmpFatalErrorException' has no required constructor"));
		}
	}
	else
	{
		ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: No class 'com/nextian/icmpping/jni/exception/IcmpFatalErrorException'"));
	}
}

/*
 * Set attribute of provided Java object. The attribute type must be of type <i>byte</i>.
 */
static BOOL setByteAttribute(JNIEnv *env, jclass clazz, jobject obj, const char *fieldName, INT8 value)
{
	jfieldID fieldId;
	BOOL result;

	fieldId = (*env)->GetFieldID(env, clazz, fieldName, "B");
	if (fieldId != NULL)
	{
		ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: Have fieldId for byte field %s"), fieldName);
		(*env)->SetByteField(env, obj, fieldId, value);
		ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: set %s"), fieldName);
		result = TRUE;
	}
	else result = FALSE;
	return result;
}

/*
 * Set attribute of provided Java object. The attribute type must be of type <i>short</i>.
 */
static BOOL setShortAttribute(JNIEnv *env, jclass clazz, jobject obj, const char *fieldName, INT16 value)
{
	jfieldID fieldId;
	BOOL result;

	fieldId = (*env)->GetFieldID(env, clazz, fieldName, "S");
	if (fieldId != NULL)
	{
		ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: Have fieldId for short field %s"), fieldName);
		(*env)->SetShortField(env, obj, fieldId, value);
		ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: set %s"), fieldName);
		result = TRUE;
	}
	else result = FALSE;
	return result;
}

/*
 * Set attribute of provided Java object. The attribute type must be of type <i>int</i>.
 */
static BOOL setIntAttribute(JNIEnv *env, jclass clazz, jobject obj, const char *fieldName, INT32 value)
{
	jfieldID fieldId;
	BOOL result;

	fieldId = (*env)->GetFieldID(env, clazz, fieldName, "I");
	if (fieldId != NULL)
	{
		ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: Have fieldId for int field %s"), fieldName);
		(*env)->SetIntField(env, obj, fieldId, value);
		ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: set %s"), fieldName);
		result = TRUE;
	}
	else result = FALSE;
	return result;
}

/*
 * Set attribute of provided Java object. The attribute type must be of type <i>byte[]</i>.
 */
static BOOL setByteArrayAttribute(JNIEnv *env, jclass clazz, jobject obj, const char *fieldName, INT8 *arrayBuffer, long arrayLength)
{
	jfieldID fieldId;
	jbyteArray byteArray;
	BOOL result;

	if (arrayLength > 0)
	{
		fieldId = (*env)->GetFieldID(env, clazz, fieldName, "[B");
		if (fieldId != NULL)
		{
			ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: Have fieldId for short field %s"), fieldName);
			byteArray = (*env)->NewByteArray(env, arrayLength);
			if (byteArray != NULL)
			{
				ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Have new array"));
				(*env)->SetByteArrayRegion(env, byteArray, 0, arrayLength, arrayBuffer);
				ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Have array created and set"));
				(*env)->SetObjectField(env, obj, fieldId, byteArray);
				ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: set %s"), fieldName);
				result = TRUE;
			}
			else result = FALSE;
		}
		else result = FALSE;
	}
	else result = TRUE;
	return result;
}

/*
 * Set attribute of provided Java object. The attribute type must be of type <i>InetAddress</i>.
 */
static BOOL setInetAddressAttribute(JNIEnv *env, jclass clazz, jobject obj, const char *fieldName, UINT8 *address)
{
	int addressLen = 4;
	jclass inet4AddressClass;
	jobject addressObj;
	jmethodID getByAddressMethodId;
	jfieldID fieldId;
	jbyteArray byteArray;
	BOOL result;

	fieldId = (*env)->GetFieldID(env, clazz, fieldName, "Ljava/net/InetAddress;");
	if (fieldId != NULL)
	{
		ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: Have fieldId for short field %s"), fieldName);
		byteArray = (*env)->NewByteArray(env, addressLen);
		if (byteArray != NULL)
		{
			ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Have new array"));
			(*env)->SetByteArrayRegion(env, byteArray, 0, addressLen, (jbyte *)address);
			// ICMPOUTPUTDEBUGSTRING0("ICMP JNI: Have array created and set");
			//(*env)->SetObjectField(env, obj, fieldId, byteArray);
			// ICMPOUTPUTDEBUGSTRING0("ICMP JNI: set %s", fieldName);
			inet4AddressClass = (*env)->FindClass(env, "java/net/InetAddress");
			if (inet4AddressClass != NULL)
			{
				ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Have InetAddress"));
				getByAddressMethodId = (*env)->GetStaticMethodID(env, inet4AddressClass, "getByAddress", "([B)Ljava/net/InetAddress;");
				if (getByAddressMethodId != NULL)
				{
					ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Have getByAddress"));
					addressObj = (*env)->CallStaticObjectMethod(env, inet4AddressClass, getByAddressMethodId, byteArray);
					if (addressObj != NULL)
					{
						ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Have address object"));
						(*env)->SetObjectField(env, obj, fieldId, addressObj);
						result = TRUE;
					}
					else
					{
						ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: could not create address"));
						result = FALSE;
					}
				}
				else
				{
					ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: no getByAddress method"));
					result = FALSE;
				}
			}
			else
			{
				ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: no Inet4Address class"));
				result = FALSE;
			}
		}
		else
		{
			ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: no byte array"));
			result = FALSE;
		}
	}
	else
	{
		ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: no field %s"), fieldName);
		result = FALSE;
	}
	return result;
}

/*
 * Java interface to IcmpCreateSocket function. In case of error returned,
 * the IcmpSystemException is thrown.
 */
JNIEXPORT jlong JNICALL JNICALL Java_com_nextian_icmpping_jni_IcmpSocket_nativeCreateSocket(JNIEnv *env, jobject obj)
{
	SOCKET socket = IcmpCreateSocket();

	if (socket == SOCKET_ERROR) throwSystemException(env, IcmpGetLastError());
	return (long)socket;
	UNREFERENCED_PARAMETER(obj);
}

/*
 * Java interface to IcmpCloseSocket function. In case of error returned,
 * the IcmpSystemException is thrown.
 */
JNIEXPORT void JNICALL Java_com_nextian_icmpping_jni_IcmpSocket_nativeCloseSocket(JNIEnv *env, jclass obj, jlong handle)
{
	int result = IcmpCloseSocket((SOCKET)handle);

	if (result == SOCKET_ERROR) throwSystemException(env, IcmpGetLastError());
	UNREFERENCED_PARAMETER(obj);
}

/*
 * Java interface to IcmpWait function. In case of error returned,
 * the IcmpSystemException is thrown, when timeout occurs IcmpTimeoutException is thrown,
 * in case of success (data ready to read) function passes through.
 */
JNIEXPORT void JNICALL Java_com_nextian_icmpping_jni_IcmpSocket_nativeSelectSocket(JNIEnv *env, jobject obj, jlong handle, jint timeout)
{
	int result = IcmpWait((SOCKET)handle, timeout);

	if (result == SOCKET_ERROR)
		throwSystemException(env, IcmpGetLastError());
	else if (result == 0)
		throwTimeoutException(env);
	UNREFERENCED_PARAMETER(obj);
}

/*
 * Java interface to IcmpSendEcho function. Input messageId variable (32 bits) is split
 * into two 16bits variables that are assigned to ICMP echo identifier and sequenceNumber.
 * In case of error returned, the IcmpSystemException is thrown, in case of success function returns
 * size of sent message.
 */
JNIEXPORT jint JNICALL Java_com_nextian_icmpping_jni_IcmpSocket_nativeSendEcho(JNIEnv *env, jobject obj, jlong handle, jbyteArray destAddress, jint messageId, jbyteArray dataBuffer, jint dataLength)
{
	SOCKET socket = (SOCKET)handle;
	jbyte *dataBuf = NULL;
	jbyte *addrBuf = NULL;
	UINT16 identifier;
	UINT16 sequenceNumber;
	int result;

	addrBuf = (*env)->GetByteArrayElements(env, destAddress, NULL);
	dataBuf = (*env)->GetByteArrayElements(env, dataBuffer, NULL);
	identifier = (UINT16)((messageId >> 16) & 0xFFFF);
	sequenceNumber = (UINT16)(messageId & 0xFFFF);

	result = IcmpSendEcho(socket, addrBuf, identifier, sequenceNumber, dataBuf, dataLength);
	if (result == SOCKET_ERROR)
		throwSystemException(env, IcmpGetLastError());

	(*env)->ReleaseByteArrayElements(env, destAddress, addrBuf, JNI_ABORT);
	(*env)->ReleaseByteArrayElements(env, dataBuffer, dataBuf, JNI_ABORT);
	return result;
	UNREFERENCED_PARAMETER(obj);
}

/*
 * Java interface to IcmpReadMessage. The received message is decoded and
 * new IcmpEchoResponse object containing decoded data is returned. In case of error returned,
 * the IcmpSystemException is thrown, in case of error in object creation the IcmpFatalErrorException
 * is thrown.
 */
JNIEXPORT jobject JNICALL Java_com_nextian_icmpping_jni_IcmpSocket_nativeReceiveMessage(JNIEnv *env, jobject obj, jlong handle)
{
	SOCKET socket = (SOCKET)handle;
	IcmpResponseMessage response;
	UINT8  srcAddress[4];
	UINT8  dstAddress[4];
	UINT8  type;
	UINT8  code;
	UINT16 identifier;
	UINT16 sequenceNumber;
	UINT32 messageId;
	INT8  *dataBuffer;
	long dataLength;
	jclass icmpEchoResponseClass;
	jmethodID icmpEchoResponseConstructorId;
	jobject icmpEchoResponseObject;
	BOOL setResult;
	jobject result = NULL;

	if (IcmpReadMessage(socket, &response) != SOCKET_ERROR)
	{
		dataBuffer = malloc(sizeof(INT8) * ICMP_ECHO_MAX_DATA_LEN);
		if (IcmpDecodeResponseMessage(&response, srcAddress, dstAddress, &type, &code, &identifier, &sequenceNumber, dataBuffer, ICMP_ECHO_MAX_DATA_LEN, &dataLength) == TRUE)
		{
			ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Got response"));
			icmpEchoResponseClass = (*env)->FindClass(env, "com/nextian/icmpping/jni/IcmpEchoResponse");
			if (icmpEchoResponseClass != NULL)
			{
				ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Found class"));
				icmpEchoResponseConstructorId = (*env)->GetMethodID(env, icmpEchoResponseClass, "<init>", "()V");
				if (icmpEchoResponseConstructorId != NULL)
				{
					ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Found constructor"));
					icmpEchoResponseObject = (*env)->NewObject(env, icmpEchoResponseClass, icmpEchoResponseConstructorId);
					if (icmpEchoResponseObject != NULL)
					{
						messageId = ((UINT32)(identifier) << 16) + sequenceNumber;
						ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: Got response id %d"), messageId);
						setResult = setIntAttribute(env, icmpEchoResponseClass, icmpEchoResponseObject, "messageId", messageId);
						if (setResult) setResult = setByteAttribute(env, icmpEchoResponseClass, icmpEchoResponseObject, "type", type);
						if (setResult) setResult = setByteAttribute(env, icmpEchoResponseClass, icmpEchoResponseObject, "code", code);
						if (setResult) setResult = setByteArrayAttribute(env, icmpEchoResponseClass, icmpEchoResponseObject, "data", dataBuffer, dataLength);
						if (setResult) setResult = setInetAddressAttribute(env, icmpEchoResponseClass, icmpEchoResponseObject, "sourceAddress", srcAddress);
						if (setResult) setResult = setInetAddressAttribute(env, icmpEchoResponseClass, icmpEchoResponseObject, "destinationAddress", dstAddress);
						if (setResult)
							result = icmpEchoResponseObject;
						else
							throwFatalErrorEception(env, JNI_ERR_COULD_NOT_SET_FIELD);
						ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: go for result"));
					}
					else
					{
						ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Could not create exception object"));
						throwFatalErrorEception(env, JNI_ERR_COULD_NOT_CREATE_RESPONSE_OBJECT);
					}
				}
				else
				{
					ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: Class 'com/nextian/icmpping/jni/IcmpEchoResponse' has no required constructor"));
					throwFatalErrorEception(env, JNI_ERR_COULD_NOT_CREATE_RESPONSE_OBJECT);
				}
			}
			else
			{
				ICMPOUTPUTDEBUGSTRING0(_T("ICMP JNI: No class 'com/nextian/icmpping/jni/IcmpEchoResponse'"));
				throwFatalErrorEception(env, JNI_ERR_COULD_NOT_CREATE_RESPONSE_OBJECT);
			}
		}
		else throwSystemException(env, IcmpGetLastError());
		free(dataBuffer);
	}
	else throwSystemException(env, IcmpGetLastError());
	return result;
	UNREFERENCED_PARAMETER(obj);
}

/*
 * Java interface to access mesageIdentifier counter.
 * On call it returns messageIdentifier value and increased it.
 * The global (opened by all applications using this module) mutex is used to
 * synchronize messageIdentifier access.
 */
JNIEXPORT jint JNICALL Java_com_nextian_icmpping_jni_IcmpSocket_nativeGetNextMessageId(JNIEnv *env, jobject obj)
{
	jint result;

	IcmpLockMutex(mutex);
	result = icmpMessageIdentifier;
	if (icmpMessageIdentifier == INT_MAX)
		icmpMessageIdentifier = 0;
	else
		icmpMessageIdentifier += 1;
	ICMPOUTPUTDEBUGSTRING1(_T("ICMP JNI: Get ID %d"), result);
	IcmpUnlockMutex(mutex);
	return result;
	UNREFERENCED_PARAMETER(env);
	UNREFERENCED_PARAMETER(obj);
}
