/*
Copyright (c) Nextian. All rights reserved.

This software is furnished under a license. Use, duplication,
disclosure and all other uses are restricted to the rights
specified in the written license agreement.
*/

#ifndef __ICMP_H__
#define __ICMP_H__

#include <wtypes.h>

#ifdef __cplusplus
extern "C" {
#endif

// Pack data structures so no data padding is used in messages/packets definitions
#pragma pack(1)

// Misc ICMP constants
#define IP_PROTOCOL_ICMP             1
#define ICMP_ECHOREPLY               0
#define ICMP_DESTINATIONUNREACHABLE  3
#define ICMP_SOURCEQUENCH            4
#define ICMP_REDIRECT                5
#define ICMP_ECHOREQEST              8
#define ICMP_TIMEEXCEEDED           11
#define ICMP_PARAMETERPROBLEM       12

/// <summary>IP header length in bytes</summary>
#define IP_HEADER_LEN        20     

/// <summary>Minimum length of ICMP payload message (may be followed by additional data)</summary>
#define ICMP_PAYLOAD_MIN_LEN 8

/// <summary>Length of ICMP error message</summary>
#define ICMP_ERROR_MESSAGE_LEN (IP_HEADER_LEN + ICMP_PAYLOAD_MIN_LEN)

/// <summary>Maximum size of an IP packet</summary>
#define IP_MESSAGE_MAX_LEN   0xFFFF 

/// <summary>Maximum size of ICMP echo data content</summary>
#define ICMP_ECHO_MAX_DATA_LEN  (IP_MESSAGE_MAX_LEN - ICMP_PAYLOAD_MIN_LEN - IP_HEADER_LEN)

	/// <summary>IP Header as defined in RFC 791</summary>
	typedef struct tagIpHeader
	{
		UINT8 versionIhl;                   // version, IHL (compressed on one byte)
		UINT8 tos;                          // type of service
		UINT16 length;                      // total length
		UINT16 id;                          // identification
		UINT16 flagsOffset;                 // flags, fragment offset (compressed on two bytes)
		UINT8 ttl;                          // time to live (TTL)
		UINT8 protocol;                     // protocol
		UINT16 checksum;                    // header checksum
		struct in_addr iaddrSrc;            // source address
		struct in_addr iaddrDst;            // destination address
	} IpHeader;
	
	/// <summary>ICMP Redirect Quench as defined in RFC 792</summary>
	typedef struct tagIcmpRedirectQuench
	{
		struct in_addr gatewayAddr;         // gateway address
	} IcmpRedirectQuench;

	/// <summary>ICMP Echo Request or Echo Response quench as defined in RFC 792</summary>
	typedef struct tagIcmpEchoQuench
	{
		UINT16 id;                          // identifier
		UINT16 seq;                         // sequence number
	} IcmpEchoQuench;

	/// <summary>ICMP Quench (redirect or echo)</summary>
	typedef union tagIcmpQuench
	{
		IcmpRedirectQuench redirect;
		IcmpEchoQuench echo;
	} IcmpQuench;

	/// <summary>Common ICMP message header as defined in RFC 792</summary>
	typedef struct tagIcmpHeader
	{
		UINT8 type;
		UINT8 code;
		UINT16 checksum;
		IcmpQuench quench;
	} IcmpHeader;

	/// <summary>ICMP response data. According to ICMP specification, ICMP response message 
	/// contains additional data. This data includes IP message header and 16 bytes of IP 
	/// message data. In this implementation, response always contains 16 bytes of 
	/// ICMP message that is IcmpHeader structure.</summary>
	typedef struct tagIcmpRawMessage
	{
		IpHeader ipHeader;		// header
		IcmpHeader icmp;		// payload
	} IcmpRawMessage;

	/// <summary>ICMP response data. The data may be at most ICMP_ECHO_MAX_DATA_LEN bytes long.
	/// In this implementation, data content of an ICMP message contains either raw data 
	/// (this may be fetched from buffer field) or structured ICMP raw message (this includes 
	/// IP message header).</summary>
	typedef union tagIcmpData
	{
		IcmpRawMessage packet;
		UINT8 buffer[ICMP_ECHO_MAX_DATA_LEN]; // ICMP data placeholder (the real data size is calculated from IP header length)
	} IcmpData;

	/// <summary>ICMP Message as defined in RFC 792. It contains ICMP header (type of request),
	/// request specific attributes (quench), and optional message data.</summary>
	typedef struct tagIcmpMessage
	{
		IcmpHeader header;
		IcmpData data;
	} IcmpMessage;

	/// <summary>ICMP response message received from Win32 raw socket, containing IP header 
	/// and ICMP message as defined in RFC 792.</summary>
	typedef struct tagIcmpResponseMessage
	{
		IpHeader ipHeader;
		IcmpMessage icmp;
	} IcmpResponseMessage;

#pragma pack()

	/// <summary>Initialize ICMP and sockets (WSAStartup()). Call IcmpRelease() after use to release 
	/// all resources allocated by this function.</summary>
	/// <returns>0 on success, other WSAStartup codes on failure</returns>
	/// <seealso cref = "IcmpRelease()" />
	int IcmpInitialize();

	/// <summary>Release resources allocated by IcmpInitialize()</summary>
	/// <seealso cref = "IcmpInitialize()" />
	void IcmpRelease();

	/// <summary>Create a raw socket so it can be used for ICMP. Call IcmpCreateSocket() after use to close socket and release system resources.
	/// This function requires <b>elevated privileges</b>.</summary>
	/// <returns>created socket handle or SOCKET_ERROR upon error</returns>
	SOCKET IcmpCreateSocket();

	/// <summary>Close socket freeing associated system resources (must be closed for each socket opened via IcmpSocket()). Function is thread safe.</summary>
	/// <param name="socket">raw socket handle to close (obtained by IcmpCreateSocket())</param>
	/// <returns>SOCKET_ERROR in case of error, other value in success</returns>
	int IcmpCloseSocket(SOCKET socket);

	/// <summary>Wait until socket has data to read</summary>
	/// <param name="socket">socket to wait for data</param>
	/// <param name="timeout">wait timeout in milliseconds</param>
	/// <returns>0 on timeout, 1 when data is ready or SOCKET_ERROR in case of error (e.g. socket has been closed in other thread)</returns>
	int IcmpWait(SOCKET socket, long timeout);

	/// <summary>Asynchronously send ICMP Echo request.</summary>
	/// <param name="socket">socket to send the message through</param>
	/// <param name="destAddress">address to send message to</param>
	/// <param name="identifier">unique identifier of the message</param>
	/// <param name="sequenceNumber">sequence number of the message</param>
	/// <param name="dataBuffer">buffer containing data to be sent</param>
	/// <param name="dataLength">length of data to be sent</param>
	/// <returns>number of bytes sent on success or SOCKET_ERROR on failure</returns>
	int IcmpSendEcho(SOCKET socket, const INT8 *destAddress, UINT16 identifier, UINT16 sequenceNumber, const void *dataBuffer, long dataLength);

	/// <summary>Read message from socket assuming received message has ICMP structure (this is however not verified).
	/// The function does not change socket mode (blocking/not blocking) so it is up to a developer to 
	/// properly handle socket behavior.</summary>
	/// <param name="socket">socket to read message from</param>
	/// <param name="response">response message pointer</param>
	/// <returns>size of message received or SOCKET_ERROR on failure</returns>
	int IcmpReadMessage(SOCKET socket, IcmpResponseMessage *response);

	/// <summary>Decode ICMP response message</summary>
	/// <param name="message"> message to be decoded</param>
	/// <param name="srcAddress">message sender address</param>
	/// <param name="dstAddress">ICMP EchoRequest destination address</param>
	/// <param name="type">ICMP message type</param>
	/// <param name="code">message code (related to received message type, see RFC 791)</param>
	/// <param name="identifier">message identifier</param>
	/// <param name="sequenceNumber">message sequence number</param>
	/// <param name="dataBuffer">data buffer to store decoded ICMP EchoResponse data</param>
	/// <param name="dataBuffferLength">maximum length of the data buffer</param>
	/// <param name="dataLength">length of received data</param>
	/// <returns>TRUE on success or FALSE otherwise</returns>
	BOOL IcmpDecodeResponseMessage(const IcmpResponseMessage *message, UINT8 *srcAddress, UINT8 *dstAddress, UINT8 *type, UINT8 *code, UINT16 *identifier, UINT16 *sequenceNumber, void *dataBuffer, size_t dataBuffferLength, long *dataLength);

	/// <summmary>Get error code associated with a socket using getsockopt() and SO_ERROR</summary>
	/// <param name="socket">socket to retrieve error code for</param>
	/// <returns>last socket error code or WSAEINVAL when invalid socket has been specified</returns>
	int IcmpGetSocketError(SOCKET socket);

	/// <summmary>Return Win32 API error code for the last ICMP operation</summary>
	/// <returns>error code</returns>
	/// <seealso cref = "IcmpGetErrorString()" / >
	int IcmpGetLastError();

	/// <summmary>Get description string for an error code (must be released using IcmpReleaseErrorString() after use)</summary>
	/// <param name="errorCode">error code to be described</param>
	/// <returns>error string (must be released with IcmpReleaseErrorString())</returns>
	/// <seealso cref = "IcmpReleaseErrorString()" / >
	LPTSTR IcmpGetErrorString(int errorCode);

	/// <summary>Release memory for error string allocated with IcmpReleaseErrorString</summary>
	/// <param name="errorMessage">string to release</param>
	/// <seealso cref = "IcmpReleaseErrorString()" / >
	void IcmpReleaseErrorString(LPTSTR errorMessage);

#ifdef __cplusplus
}
#endif

#endif
