/*
Copyright (c) Nextian. All rights reserved.

This software is furnished under a license. Use, duplication,
disclosure and all other uses are restricted to the rights
specified in the written license agreement.
*/


#include <tchar.h>
#include <winsock.h>
#include <mstcpip.h>
#include "CppUnitTest.h"

#include "Icmp.h"

using namespace Microsoft::VisualStudio::CppUnitTestFramework;

#define ZERO_OBJECT(o) memset(&o, 0, sizeof(o))

namespace WindowsIcmpTest
{
	/// Test suite pinging actual hosts
	TEST_CLASS(EchoTest)
	{
	private:
		/// Always reachable local address
		const char * const LOCAL_ADDRESS = "127.0.0.1";

		/// Non answering IP address
		const char * const NO_ANSWER_IP = "2.3.4.5";

		/// Unreachable IP address (most likely has to be changed for tests - unreachable address is a non-responding 
		/// address on your network)
		const char * const UNRECHABLE_IP = "10.0.0.50";

		const UINT16 ICMP_ID = 123;

		const UINT16 ICMP_SEQ = 987;

		/// Test message sent in ICMP packets (ASCII string)
		const void * const ICMP_DATA = "Windows ICMP test";

		/// Length of test message in octet/bytes
		const int ICMP_DATA_LEN = lstrlenA(static_cast<const char *>(ICMP_DATA));

		/// Maximum wait timeout (milliseconds)
		const int maxWaitTimeout = 5000;

	private:
		void inline SetLocalhostAddress(struct sockaddr_in* dstAddr)
		{
			ZERO_OBJECT(*dstAddr);
			dstAddr->sin_addr.s_addr = inet_addr(LOCAL_ADDRESS);
		}

		void inline SetNoanswerAddress(struct sockaddr_in* dstAddr)
		{
			ZERO_OBJECT(*dstAddr);
			dstAddr->sin_addr.s_addr = inet_addr(NO_ANSWER_IP);
		}

		void inline SetUnreachableAddress(struct sockaddr_in* dstAddr)
		{
			ZERO_OBJECT(*dstAddr);
			dstAddr->sin_addr.s_addr = inet_addr(UNRECHABLE_IP);
		}

	public:

		/// Initialize sockets for tests
		TEST_CLASS_INITIALIZE(StartupSockets)
		{
			WSADATA wsaData;
			int iResult = WSAStartup(MAKEWORD(2, 2), &wsaData);
			Assert::IsTrue(iResult == 0);
		}

		/// Shutdown sockets after tests
		TEST_CLASS_CLEANUP(ShutdownSockets)
		{
			int iResult = WSACleanup();
			Assert::IsTrue(iResult == 0);
		}

		/// Send echo, not waiting for response.
        ///  Test sending ICMP packet. Operation should always succeed unless
        ///  there are system level errors (e.g. missing permissions).
        ///  During this test we do not expect any response message.
		TEST_METHOD(SendEcho)
		{
			SOCKET socket = IcmpCreateSocket();
			Assert::IsTrue(socket != SOCKET_ERROR);

			struct sockaddr_in dstAddr;
			SetLocalhostAddress(&dstAddr);

			int iResult = IcmpSendEcho(
				socket,
				(INT8 *)&dstAddr.sin_addr.S_un.S_addr,
				ICMP_ID,
				ICMP_SEQ,
				ICMP_DATA,
				ICMP_DATA_LEN);
			Assert::AreNotEqual(iResult, SOCKET_ERROR);

			Assert::AreNotEqual(IcmpCloseSocket(socket), SOCKET_ERROR);
		}

		/// Ping localhost.
        ///  Test sending ICMP packet and verify proper response has been retrieved.
        ///  Received response should:
        ///   - contain the same message identifier and sequence number as echo request
        ///   - message type should be ICMP_ECHOREPLY with code 0
        ///   - data should be equal to the sent one
        ///   - destination address should match
		TEST_METHOD(WaitForAndCheckResponse)
		{
			IcmpResponseMessage response;
			struct sockaddr_in dstAddr;
			UINT8  srcAddress[4];
			UINT8  dstAddress[4];
			UINT8  type;
			UINT8  code;
			UINT16 identifier;
			UINT16 sequenceNumber;
			INT8   dataBuffer[ICMP_ECHO_MAX_DATA_LEN];
			long dataLength;

			SOCKET socket = IcmpCreateSocket();
			Assert::IsTrue(socket != SOCKET_ERROR);

			SetLocalhostAddress(&dstAddr);

			int iResult = IcmpSendEcho(socket, (INT8 *)&dstAddr.sin_addr.S_un.S_addr, ICMP_ID, ICMP_SEQ, ICMP_DATA, ICMP_DATA_LEN);
			Assert::AreNotEqual(iResult, SOCKET_ERROR);

			Assert::AreEqual(IcmpWait(socket, maxWaitTimeout), 1 /* 1 response returned */);
			Assert::AreNotEqual(IcmpReadMessage(socket, &response), SOCKET_ERROR);
			Assert::AreNotEqual(IcmpCloseSocket(socket), SOCKET_ERROR);

			Assert::AreEqual(IcmpDecodeResponseMessage(&response, srcAddress, dstAddress, &type, &code, &identifier, &sequenceNumber, dataBuffer, ICMP_ECHO_MAX_DATA_LEN, &dataLength), TRUE);
			Assert::IsTrue(identifier == ICMP_ID);
			Assert::IsTrue(sequenceNumber == ICMP_SEQ);
			Assert::AreEqual(type, static_cast<UINT8>(ICMP_ECHOREPLY));
			Assert::AreEqual(code, static_cast<UINT8>(0));
			Assert::IsTrue(dataLength == ICMP_DATA_LEN);
			Assert::AreEqual(memcmp(dstAddress, &dstAddr.sin_addr.S_un.S_addr, sizeof(dstAddress)), 0);
			Assert::AreEqual(memcmp(dataBuffer, ICMP_DATA, dataLength), 0);
		}

		/// Ping a non-answering host
        ///  Test sending ICMP packet to not existing IP address (taken arbitrary)
        ///  Sending should succeed but waiting for response should fail with timeout 
		TEST_METHOD(NoAnswer)
		{
			SOCKET socket = IcmpCreateSocket();
			Assert::IsTrue(socket != SOCKET_ERROR);

			struct sockaddr_in dstAddr;
			SetNoanswerAddress(&dstAddr);

			int iResult = IcmpSendEcho(socket, (INT8 *)&dstAddr.sin_addr.S_un.S_addr, ICMP_ID, ICMP_SEQ, ICMP_DATA, ICMP_DATA_LEN);
			Assert::AreNotEqual(iResult, SOCKET_ERROR);

			Assert::AreEqual(IcmpWait(socket, maxWaitTimeout), 0 /* 0 -> timeout */);
			Assert::AreNotEqual(IcmpCloseSocket(socket), SOCKET_ERROR);
		}

		/// Ping an unreachable host
        ///  Test sending ICMP packet to address reported by IP stack as unreachable.
        ///  Sending should succeed and response should be received.
        ///  The response should:
        ///   - contain the same message identifier and sequence number as echo request
        ///   - message type should be ICMP_DESTINATIONUNREACHABLE
        ///   - destination address should match
		///
		/// CAUTION: This test may require of destination addressing changing (see UNRECHABLE_IP constant)
		///
		TEST_METHOD(Unreachable)
		{
			IcmpResponseMessage response;
			struct sockaddr_in dstAddr;
			UINT8  srcAddress[4];
			UINT8  dstAddress[4];
			UINT8  type;
			UINT8  code;
			UINT16 identifier;
			UINT16 sequenceNumber;
			INT8   dataBuffer[ICMP_ECHO_MAX_DATA_LEN];
			long dataLength;

			SOCKET socket = IcmpCreateSocket();
			Assert::IsTrue(socket != SOCKET_ERROR);

			SetUnreachableAddress(&dstAddr);

			int iResult = IcmpSendEcho(socket, (INT8 *)&dstAddr.sin_addr.S_un.S_addr, ICMP_ID, ICMP_SEQ, ICMP_DATA, ICMP_DATA_LEN);
			Assert::AreNotEqual(iResult, SOCKET_ERROR);

			Assert::AreEqual(IcmpWait(socket, maxWaitTimeout), 1 /* we should actually get a response */);
			Assert::AreNotEqual(IcmpReadMessage(socket, &response), SOCKET_ERROR);
			Assert::AreNotEqual(IcmpCloseSocket(socket), SOCKET_ERROR);

			Assert::AreEqual(IcmpDecodeResponseMessage(&response, srcAddress, dstAddress, &type, &code, &identifier, &sequenceNumber, dataBuffer, ICMP_ECHO_MAX_DATA_LEN, &dataLength), TRUE);
			Assert::IsTrue(identifier == ICMP_ID);
			Assert::IsTrue(sequenceNumber == ICMP_SEQ);
			Assert::AreEqual(type, static_cast<UINT8>(ICMP_DESTINATIONUNREACHABLE));
			Assert::AreEqual(memcmp(dstAddress, &dstAddr.sin_addr.S_un.S_addr, sizeof(dstAddress)), 0);
		}

		/// Wait for response without sending request, causing a timeout
		TEST_METHOD(ResponseTimeout)
		{
			SOCKET socket = IcmpCreateSocket();
			Assert::IsTrue(socket != SOCKET_ERROR);
			Assert::AreEqual(IcmpWait(socket, maxWaitTimeout), 0 /* 0 -> timeout */);
			Assert::AreNotEqual(IcmpCloseSocket(socket), SOCKET_ERROR);
		}

		/// Wait for response on an illegal socket handle.
        ///  Test IcmpWait function to properly return error on invalid socket handle parameter.
        ///  Function should not throw exception!
		TEST_METHOD(WaitOnInvalidSocket)
		{
			Assert::IsTrue(IcmpWait(static_cast<SOCKET>(-1/* invalid socket value*/), maxWaitTimeout) == SOCKET_ERROR);
		}

		/// Provide a wrong data buffer to IcmpSendEcho
        ///   Test IcmpSendEcho function to properly handle invalid parameter. If wrong data pointer
        ///   is provided, function should fail with WSAEINVAL result code. Function should not throw exception!
		TEST_METHOD(IllegalSendData)
		{
			SOCKET socket = IcmpCreateSocket();
			Assert::IsTrue(socket != SOCKET_ERROR);

			struct sockaddr_in dstAddr;
			SetLocalhostAddress(&dstAddr);

			int iResult = IcmpSendEcho(socket, (INT8 *)&dstAddr.sin_addr.S_un.S_addr, ICMP_ID, ICMP_SEQ, (char *)-1, 4);
			Assert::AreEqual(iResult, SOCKET_ERROR);
			Assert::IsTrue(IcmpGetLastError() == WSAEINVAL);

			Assert::AreNotEqual(IcmpCloseSocket(socket), SOCKET_ERROR);
		}

		/// Provide a wrong data buffer to IcmpReadMessage
		///   Test IcmpReadMessage function to properly handle invalid parameter. If wrong data pointer
		///   is provided, function should fail with WSAEINVAL result code. Function should not throw exception!
		TEST_METHOD(IllegalReceiveData)
		{
			SOCKET socket = IcmpCreateSocket();
			Assert::IsTrue(socket != SOCKET_ERROR);

			struct sockaddr_in dstAddr;
			SetLocalhostAddress(&dstAddr);

			int iResult = IcmpSendEcho(socket, (INT8 *)&dstAddr.sin_addr.S_un.S_addr, ICMP_ID, ICMP_SEQ, ICMP_DATA, ICMP_DATA_LEN);
			Assert::AreNotEqual(iResult, SOCKET_ERROR);

			Assert::AreEqual(IcmpReadMessage(socket, (IcmpResponseMessage *)NULL), SOCKET_ERROR);
			Assert::IsTrue(IcmpGetLastError() == WSAEINVAL);
			Assert::AreNotEqual(IcmpCloseSocket(socket), SOCKET_ERROR);
		}

		/// Ping illegal address
        ///  Test IcmpSendEcho function by sending ICMP packet to illegal IP address.
        ///  Function should fail with code WSAEACCES.
		TEST_METHOD(IllegalAddress)
		{
			SOCKET socket = IcmpCreateSocket();
			Assert::IsTrue(socket != SOCKET_ERROR);

			long illegalAddr = -1;

			int iResult = IcmpSendEcho(socket, (INT8 *)&illegalAddr, ICMP_ID, ICMP_SEQ, ICMP_DATA, ICMP_DATA_LEN);
			Assert::AreEqual(iResult, SOCKET_ERROR);

			Assert::IsTrue(IcmpGetLastError() == WSAEACCES);

			Assert::AreNotEqual(IcmpCloseSocket(socket), SOCKET_ERROR);
		}


	};
}
